#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <string.h>

#ifdef WIN32
#include <process.h>
#else
#include <stdlib.h>
#include <unistd.h>
#endif

#include "object.h"
#include "util.h"
#include "xmalloc.h"
#include "stringcache.h"
#include "TAStreamTokenizer.h"

/**
 * A class to turn an TA input stream into a stream of tokens.
 * @version 1.0, 22/07/98
 * @author	Gary Farmaner
 */

char *CStreamTokenizer::m_ctypeP = 0;

void 
CStreamTokenizer::dumpBuf(int lth)
{
	char	*bufP = m_bufP;
	int		i;

	printf("(%d) ", lth);
	for (i = 0; i < lth && i < 30; ++i) {
		fputc(bufP[i], stderr);
	}
	if (lth > 30) {
		fprintf(stderr, " ... ");
		i = lth - 30;
		if (i < 30) {
			i = 30;
		}
		for (; i < lth; ++i) {
			fputc(bufP[i], stderr);
		}
}	}

int 
CStreamTokenizer::enlargeBuffer()
{
	int		need     = m_buf_size * 2;

	m_bufP = (char *) Xrealloc(m_bufP, need);
	m_buf_size = need; 
	return(need);
}

const char *
CStreamTokenizer::getString(int pos)
{
	if (pos >= m_buf_size) {
		enlargeBuffer();
	}
	m_bufP[pos] = 0;
	return stringCache(m_bufP);
}

void
CStreamTokenizer::open(const char *filenameP)
{
	if (!filenameP) {
		m_filenameP = "";
	} else {
		m_filenameP = filenameP;
	}
	if (filenameP && *filenameP) {
		m_F = fopen(filenameP, "rb");
		if (!m_F) {
			fprintf(stderr, "Can't open %s\n", filenameP);
			error_exit(2);
		}
	} else {
		m_F = stdin;
	}
	m_lineno = 1;
}

void 
CStreamTokenizer::close()
{
	if (m_F != stdin) {
		if (fclose(m_F)) {
			fprintf(stderr, "Can't close %s\n", m_filenameP);
			exit(1);
	}	}
	m_F = 0;
}

CStreamTokenizer::CStreamTokenizer(void) 
{
	int	need;

	m_F          = 0;
	m_filenameP  = 0;
	m_pushedBack = 0;
	m_escaped    = 0;
	m_svalP      = 0;
	m_lineno     = 0;

	if (!m_ctypeP) {
		m_ctypeP = (char *) Xmalloc(256);
		memset(m_ctypeP, 0, 256);
		m_ctypeP[' ']  = CT_DELIMIT;
		m_ctypeP['\f'] = CT_DELIMIT;
		m_ctypeP['\t'] = CT_DELIMIT;
		m_ctypeP['\r'] = CT_DELIMIT;
		m_ctypeP['\n'] = CT_DELIMIT;
		m_ctypeP['=']  = CT_DELIMIT;
//		m_ctypeP[':']  = CT_DELIMIT;
		m_ctypeP['{']  = CT_DELIMIT;
		m_ctypeP['}']  = CT_DELIMIT;
		m_ctypeP['(']  = CT_DELIMIT;
		m_ctypeP[')']  = CT_DELIMIT;
//		m_ctypeP['"']  = CT_QUOTE;
//		m_ctypeP['\''] = CT_QUOTE;
	}
	need        = 20 * 1024;
	m_bufP      = (char *) Xmalloc(need);		// The area that captures a token or attribute value
	m_buf_size  = need;							
	m_peekc     = ' ';
	m_ttype     = ' ';
}

int 
CStreamTokenizer::charToken() 
{
	int	c;

	c         = m_peekc;
	if (c == '\n') {
		++m_lineno;
	}
	m_peekc   = getc(m_F);
//	fputc(m_peekc, stderr);
	switch (c) {
	case '\\':
		switch (m_peekc) {
		case 'n':
			c = '\n';
			break;
		case 't':
			c = '\t';
			break;
		case 'f':
			c = '\f';
			break;
		case 'r':
			c = '\r';
			break;
		case 'e':
			c = 27;
			break;
		case 'd':
			c = 127;
		case '\\':
			m_peekc = ' ';
			break;
		case '"':
			c = '"';
			m_escaped = true;
			break;
		case '\'':
			c = '\'';
			m_escaped = true;
			break;
		default:
			c = m_peekc;
		}
		charToken();	// Consume and discard m_peekc
		break;
	case '\'':
	case '"':
		m_escaped = false;
	}
	return c;
}

/** Does an unget of the last token */

void 
CStreamTokenizer::pushBack(int ttype) 
{
	m_ttype      = ttype;
	m_pushedBack = true;
} 

const char *
CStreamTokenizer::filename()
{
	return m_filenameP;
}

/** Return the current line number. */

int 
CStreamTokenizer::lineno() 
{
	return(m_lineno);
}

/**
 * Parses a token from the input stream.  The return value is
 * the same as the value of ttype.	Typical clients of this
 * class first set up the syntax tables and then sit in a loop
 * calling nextToken to parse successive tokens until TT_EOF
 * is returned.
 */

int 
CStreamTokenizer::nextToken()
{
	if (m_pushedBack) {
		m_pushedBack = false;
		return m_ttype;
	}

	int		buf_size = m_buf_size;
	char	*bufP    = m_bufP;
	int		pos      = 0;
	int		c;

	for (m_svalP = 0; ; ) {
		// This loop is repeated while initial whitespace
		c = charToken();

		switch(c) {
		case TT_EOF:
			m_svalP = stringCache("EOF");
			return TT_EOF;
		case '\f':
		case '\r':
		case '\n':
		case '\t':
		case ' ':
			continue;
		case '"':
			for (;;) {
				c = charToken();
				switch (c) {
				case TT_EOF:
					printf("Unterminated \" in input\n");
					break;
				case '"':
					if (!m_escaped) {
						break;
					}
				default:
					if (pos >= buf_size) {
						buf_size = enlargeBuffer();
						bufP      = m_bufP;
					}	
					bufP[pos++] = (char) c;
					continue;
				}
				break;
			} 
			m_svalP = getString(pos);
			return TT_WORD;
		case '\'':
			for (;;) {
				c = charToken();
				switch (c) {
				case TT_EOF:
					printf("Unterminated ' in input\n");
					break;
				case '\'':
					if (!m_escaped) {
						break;
					}
				default:	
					if (pos >= buf_size) {
						buf_size = enlargeBuffer();
						bufP     = m_bufP;
					}	
					bufP[pos++] = (char) c;
					continue;
				}
				break;
			} 
			m_svalP = getString(pos);
/*
			if (pos < 30) {
				printf("Quote: %s\n", m_svalP);
			} else {
				printf"Quote: %30.30s ...\n" + m_svalP);
			}
*/
			return TT_WORD;
		case '/':
			switch (m_peekc) {
				case '/':
				{
					// Toss the remainder of the line
					for (;;) {
						c = charToken();
						switch (c) {
						case '\r':
							if (m_peekc == '\n') {
								continue;
							}
						case '\n':
						case TT_EOF:
							break;
						default:
							continue;
						}
						break;
					}
					continue;
				}
				case '*':
				{
					// Toss the remainder of the comment
					for (;;) {
						c = charToken();
						switch (c) {
						case '*':
							if (m_peekc != '/') {
								continue;
							}
							c = charToken();
						case TT_EOF:
							break;
						default:
							continue;
						}
						break;
					}
					continue;
			}	}
		default:
			if (c < 256 && m_ctypeP[c] == CT_DELIMIT) {
				m_svalP = 0;
				return c;
			}

			// Read characters until delimitter

			for (;;) {
				if (pos >= buf_size) {
					buf_size = enlargeBuffer();
					bufP     = m_bufP;
				}	
				bufP[pos++] = (char) c;
				if (m_peekc == TT_EOF) {
					break;
				}
				if (c == '\\') {
					if (pos >= buf_size) {
						buf_size = enlargeBuffer();
						bufP     = m_bufP;
					}	
					bufP[pos++] = (char) charToken();
				} 
				if (c < 256 && m_ctypeP[m_peekc] != 0) {
					break;
				}
				c = charToken();
			}
			m_svalP = getString(pos);
			if (pos == 1 && bufP[0] == ':') {
				return ':';
			}
			return TT_WORD;
		}
	}
}

int 
CStreamTokenizer::charAVI() 
{
	int	c;

	c         = m_peekc;
	m_peekc   = getc(m_F);
//	fputc(c, stderr);
	return c;
}

const char *
CStreamTokenizer::nextAVI() 
{
	int		depth		 = 0;
	int		buf_size     = m_buf_size;
	char	*bufP        = m_bufP;
	int		pos			 = 0;
	int		tokens		 = 0;
	int		startbracket = -1;
	int		i, c, start;
	int		escaped, simple;
//	String	ret;

	for (; ; ) {
		// This loop is repeated while initial whitespace
		c = charAVI();

		switch(c) {
		case TT_EOF:
			return 0;
		case '\f':
		case '\r':
		case '\n':
		case '\t':
		case ' ':
			continue;
		case '"':
			++tokens;
			start        = pos;
			bufP[pos++]  = '"';
			escaped      = false;
			simple       = true;
			for (;;) {
				c = charAVI();
				switch (c) {
				case TT_EOF:
					fprintf(stderr, "Unterminated \" in input\n");
				case '\\':
					escaped      = !escaped;
					if (pos >= buf_size) {
						buf_size = enlargeBuffer();
						bufP     = m_bufP;
					}	
					bufP[pos++] = '\\';
					simple      = false;
					continue;
				case '"':
					if (!escaped) {
						break;
					}
				case '\'':
					simple = false;
				default:
					if (c < 256 && m_ctypeP[c] == CT_DELIMIT) {
						simple = false;
					}
					if (pos >= buf_size) {
						buf_size = enlargeBuffer();
						bufP     = m_bufP;
					}							
					bufP[pos++] = (char) c;
					escaped     = false;
					continue;
				}
				break;
			} 
			if (!simple) {
				if (pos >= buf_size) {
					buf_size = enlargeBuffer();
					bufP     = m_bufP;
				}	
				bufP[pos++] = '"';
			} else {
				// Remove the double quotes so that integers etc parse easily
				for (--pos; ++start <= pos; bufP[start-1] = bufP[start]);
			}
			break;
		case '\'':
			++tokens;
			start        = pos;
			bufP[pos++]  = '\'';
			escaped      = false;
			simple       = true;
			for (;;) {
				c = charAVI();
				switch (c) {
				case TT_EOF:
					fprintf(stderr, "Unterminated ' in input\n");
					break;
				case '\\':
					escaped = !escaped;
					if (pos >= buf_size) {
						buf_size = enlargeBuffer();
						bufP     = m_bufP;
					}	
					bufP[pos++] = '\\';
					simple      = false;
					continue;
				case '\'':
					if (!escaped) {
						break;
					}
				case '"':
					simple = false;
				default:	
					if (c < 256 && m_ctypeP[c] == CT_DELIMIT) {
						simple = false;
					}
					if (pos >= buf_size) {
						buf_size = enlargeBuffer();
						bufP     = m_bufP;
					}							
					bufP[pos++] = (char) c;
					escaped     = false;
					continue;
				}
				break;
			} 
			if (!simple) {
				if (pos >= buf_size) {
					buf_size = enlargeBuffer();
					bufP     = m_bufP;
				}							
				bufP[pos++] = '\'';
			} else {
				// Remove the single quotes so that integers etc parse easily
				for (--pos; ++start <= pos; bufP[start-1] = bufP[start]);
			}
			break;
		case '(':
			startbracket = pos;
			if (pos >= buf_size) {
				buf_size = enlargeBuffer();
				bufP     = m_bufP;
			}							
			bufP[pos++] = '(';
			if (depth != 0) {
				tokens = 2;	// Force later test to fail -- don't want to treat ((1)) as 1
			}
			++depth;
			continue;
		case ')':
			if (depth <= 0) {
				// Ignore the ')' -- treat as parse error
				fprintf(stderr, "Bad nesting of () at %d\n", lineno());
				return 0;
			} 
			switch (bufP[pos-1]) {
			case '(':
				// Omit () altogether
				fprintf(stderr, "Empty nesting of () at %d\n", lineno());
				--pos;
				break;
			case ' ':
				// Omit ' ' before ')'
				--pos;
			default:
				if (tokens != 1) {
					if (pos >= buf_size) {
						buf_size = enlargeBuffer();
						bufP     = m_bufP;
					}							
					bufP[pos++] = ')';
				} else {
					// Drop the brackets
					--pos;
					for (i = startbracket; i < pos; ++i) {
						bufP[i] = bufP[i+1];
			}	}	}
			--depth;
			break;
		case '/':
			switch (m_peekc) {
				case '/':
				{
					// Toss the remainder of the line
					for (;;) {
						c = charAVI();
						switch (c) {
						case '\r':
							if (m_peekc == '\n') {
								continue;
							}
						case '\n':
						case TT_EOF:
							break;
						default:
							continue;
						}
						break;
					}
					continue;
				}
				case '*':
				{
					// Toss the remainder of the comment
					for (;;) {
						c = charAVI();
						switch (c) {
						case '*':
							if (m_peekc != '/') {
								continue;
							}
							c = charAVI();
						case TT_EOF:
							break;
						default:
							continue;
						}
						break;
					}
					continue;
			}	}
		default:
			if (c < 256 && m_ctypeP[c] == CT_DELIMIT) {
				pushBack(c);
				return 0;
			}

			// Read characters until delimitter

			++tokens;
			start    = pos;
			simple   = true;
			for (;;) {
				if (c == '\\') {
					simple = false;
				}
				if (pos >= buf_size) {
					buf_size = enlargeBuffer();
					bufP     = m_bufP;
				}
				bufP[pos++] = (char) c;
				if (m_peekc == TT_EOF) {
					break;
				}
				if (c < 256 && m_ctypeP[m_peekc] != 0) {
					break;
				}
				c = charAVI();
			}
			if (!simple) {
				// If the string contains \ then quote it so that parseString kept simple
				if (pos >= buf_size) {
					buf_size = enlargeBuffer();
					bufP     = m_bufP;
				}							
				for (i = pos; i >= start; --i) {
					bufP[i+1] = bufP[i];
				}
				bufP[start] = '"';
				if (pos >= buf_size) {
					buf_size = enlargeBuffer();
					bufP     = m_bufP;
				}							
				bufP[pos++] = '"';
			}
			break;
		}
		if (depth > 0) {
			if (pos >= buf_size) {
				buf_size = enlargeBuffer();
				bufP     = m_bufP;
			}							
			bufP[pos++] = ' ';
			continue;
		}
		if (tokens == 0) {
			return 0;
		}

/*
		fprintf(stderr, "token :");
		dumpBuf(pos);
		fputc('\n', stderr);
*/
		return getString(pos);
	}
}

